"""Common functions used in Item Loaders code"""

from functools import partial
from typing import Any, Callable, MutableMapping

from itemloaders.utils import get_func_args


def wrap_loader_context(
    function: Callable[..., Any], context: MutableMapping[str, Any]
) -> Callable[..., Any]:
    """Wrap functions that receive loader_context to contain the context
    "pre-loaded" and expose a interface that receives only one argument
    """
    if "loader_context" in get_func_args(function):
        return partial(function, loader_context=context)
    else:
        return function



"""
Copy/paste from scrapy source at the moment, to ensure tests are working.
Refactoring to come later
"""

import inspect
from functools import partial
from typing import Any, Callable, Generator, Iterable, List


def arg_to_iter(arg: Any) -> Iterable[Any]:
    """Return an iterable based on *arg*.

    If *arg* is a list, a tuple or a generator, it will be returned as is.

    If *arg* is ``None``, an empty list will be returned.

    If *arg* is anything else, a list will be returned with *arg* as its only
    item, i.e. ``[arg]``.
    """
    if arg is None:
        return []
    if isinstance(arg, (list, tuple, Generator)):
        return arg
    return [arg]


def get_func_args(func: Callable[..., Any], stripself: bool = False) -> List[str]:
    """Return the argument name list of a callable object"""
    if not callable(func):
        raise TypeError(f"func must be callable, got {type(func).__name__!r}")

    args: List[str] = []
    try:
        sig = inspect.signature(func)
    except ValueError:
        return args

    if isinstance(func, partial):
        partial_args = func.args
        partial_kw = func.keywords

        for name, param in sig.parameters.items():
            if param.name in partial_args:
                continue
            if partial_kw and param.name in partial_kw:
                continue
            args.append(name)
    else:
        for name in sig.parameters.keys():
            args.append(name)

    if stripself and args and args[0] == "self":
        args = args[1:]
    return args



import functools
import operator
import platform
import unittest
from typing import Any

from itemloaders.utils import get_func_args


class UtilsPythonTestCase(unittest.TestCase):
    def test_get_func_args(self):
        def f1(a, b, c):
            pass

        def f2(a, b=None, c=None):
            pass

        def f3(a, b=None, *, c=None):
            pass

        class A:
            def __init__(self, a: Any, b: Any, c: Any):
                pass

            def method(self, a, b, c):
                pass

        class Callable:
            def __call__(self, a, b, c):
                pass

        a = A(1, 2, 3)
        cal = Callable()
        partial_f1 = functools.partial(f1, None)
        partial_f2 = functools.partial(f1, b=None)
        partial_f3 = functools.partial(partial_f2, None)

        self.assertEqual(get_func_args(f1), ["a", "b", "c"])
        self.assertEqual(get_func_args(f2), ["a", "b", "c"])
        self.assertEqual(get_func_args(f3), ["a", "b", "c"])
        self.assertEqual(get_func_args(A), ["a", "b", "c"])
        self.assertEqual(get_func_args(a.method), ["a", "b", "c"])
        self.assertEqual(get_func_args(partial_f1), ["b", "c"])
        self.assertEqual(get_func_args(partial_f2), ["a", "c"])
        self.assertEqual(get_func_args(partial_f3), ["c"])
        self.assertEqual(get_func_args(cal), ["a", "b", "c"])
        self.assertEqual(get_func_args(object), [])
        self.assertEqual(get_func_args(str.split, stripself=True), ["sep", "maxsplit"])
        self.assertEqual(get_func_args(" ".join, stripself=True), ["iterable"])

        if platform.python_implementation() == "CPython":
            # This didn't work on older versions of CPython: https://github.com/python/cpython/issues/86951
            self.assertIn(
                get_func_args(operator.itemgetter(2), stripself=True),
                [[], ["args", "kwargs"]],
            )
        elif platform.python_implementation() == "PyPy":
            self.assertEqual(
                get_func_args(operator.itemgetter(2), stripself=True), ["obj"]
            )


if __name__ == "__main__":
    unittest.main()



from setuptools import find_packages, setup

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="itemloaders",
    version="1.3.0",
    url="https://github.com/scrapy/itemloaders",
    project_urls={
        "Documentation": "https://itemloaders.readthedocs.io/",
        "Source": "https://github.com/scrapy/itemloaders",
    },
    description="Base library for scrapy's ItemLoader",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Zyte",
    author_email="opensource@zyte.com",
    license="BSD",
    packages=find_packages(exclude=("tests", "tests.*")),
    package_data={
        "itemadapter": ["py.typed"],
    },
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    python_requires=">=3.8",
    install_requires=[
        # before updating these versions, be sure they are not higher than
        # scrapy's requirements
        "w3lib>=1.17.0",
        "parsel>=1.5.0",
        "jmespath>=0.9.5",
        "itemadapter>=0.1.0",
    ],
    # extras_require=extras_require,
)



import unittest

from parsel import Selector

from itemloaders import ItemLoader


class SubselectorLoaderTest(unittest.TestCase):
    selector = Selector(
        text="""
    <html>
    <body>
    <header>
      <div id="id">marta</div>
      <p>paragraph</p>
    </header>
    <footer class="footer">
      <a href="http://www.scrapy.org">homepage</a>
      <img src="/images/logo.png" width="244" height="65" alt="Scrapy">
    </footer>
    </body>
    </html>
    """
    )

    def test_nested_xpath(self):
        loader = ItemLoader(selector=self.selector)
        nl = loader.nested_xpath("//header")
        nl.add_xpath("name", "div/text()")
        nl.add_css("name_div", "#id")
        assert nl.selector
        nl.add_value("name_value", nl.selector.xpath('div[@id = "id"]/text()').getall())

        self.assertEqual(loader.get_output_value("name"), ["marta"])
        self.assertEqual(
            loader.get_output_value("name_div"), ['<div id="id">marta</div>']
        )
        self.assertEqual(loader.get_output_value("name_value"), ["marta"])

        self.assertEqual(loader.get_output_value("name"), nl.get_output_value("name"))
        self.assertEqual(
            loader.get_output_value("name_div"), nl.get_output_value("name_div")
        )
        self.assertEqual(
            loader.get_output_value("name_value"), nl.get_output_value("name_value")
        )

    def test_nested_css(self):
        loader = ItemLoader(selector=self.selector)
        nl = loader.nested_css("header")
        nl.add_xpath("name", "div/text()")
        nl.add_css("name_div", "#id")
        assert nl.selector
        nl.add_value("name_value", nl.selector.xpath('div[@id = "id"]/text()').getall())

        self.assertEqual(loader.get_output_value("name"), ["marta"])
        self.assertEqual(
            loader.get_output_value("name_div"), ['<div id="id">marta</div>']
        )
        self.assertEqual(loader.get_output_value("name_value"), ["marta"])

        self.assertEqual(loader.get_output_value("name"), nl.get_output_value("name"))
        self.assertEqual(
            loader.get_output_value("name_div"), nl.get_output_value("name_div")
        )
        self.assertEqual(
            loader.get_output_value("name_value"), nl.get_output_value("name_value")
        )

    def test_nested_replace(self):
        loader = ItemLoader(selector=self.selector)
        nl1 = loader.nested_xpath("//footer")
        nl2 = nl1.nested_xpath("a")

        loader.add_xpath("url", "//footer/a/@href")
        self.assertEqual(loader.get_output_value("url"), ["http://www.scrapy.org"])
        nl1.replace_xpath("url", "img/@src")
        self.assertEqual(loader.get_output_value("url"), ["/images/logo.png"])
        nl2.replace_xpath("url", "@href")
        self.assertEqual(loader.get_output_value("url"), ["http://www.scrapy.org"])

    def test_nested_ordering(self):
        loader = ItemLoader(selector=self.selector)
        nl1 = loader.nested_xpath("//footer")
        nl2 = nl1.nested_xpath("a")

        nl1.add_xpath("url", "img/@src")
        loader.add_xpath("url", "//footer/a/@href")
        nl2.add_xpath("url", "text()")
        loader.add_xpath("url", "//footer/a/@href")

        self.assertEqual(
            loader.get_output_value("url"),
            [
                "/images/logo.png",
                "http://www.scrapy.org",
                "homepage",
                "http://www.scrapy.org",
            ],
        )

    def test_nested_load_item(self):
        loader = ItemLoader(selector=self.selector)
        nl1 = loader.nested_xpath("//footer")
        nl2 = nl1.nested_xpath("img")

        loader.add_xpath("name", "//header/div/text()")
        nl1.add_xpath("url", "a/@href")
        nl2.add_xpath("image", "@src")

        item = loader.load_item()

        assert item is loader.item
        assert item is nl1.item
        assert item is nl2.item

        self.assertEqual(item["name"], ["marta"])
        self.assertEqual(item["url"], ["http://www.scrapy.org"])
        self.assertEqual(item["image"], ["/images/logo.png"])
